
function param = ppca_md(x1, x2, feature, K, q, prior, init)

%
[N1 d] = size(x1);
N2 = size(x2,1);

% file structure
dp = feature(:,1);  % common
dq = feature(:,2);  % specific to x1
dr = feature(:,3);  % specific to x2
do1 = dp|dq;
dm1 = dr;
do2 = dp|dr;
dm2 = dq;

% constants
maxIt = 5000;       % max nunmber of iterations
tolFun = 1e-10;     % termination tolerance

% % init
if nargin < 7
    init = struct([]);
end
[pp, mu, W, sig] = util_init(x1, x2, feature, K, q, prior, init);

% 
post1 = zeros(N1, K);
post2 = zeros(N2, K);

ll_old = -Inf;
ll_hist = zeros(maxIt,1);

% % 2-step em algorithm
for iter = 1:maxIt
    
    % parameters from the previous iteration
    pp_old  = pp;
    mu_old  = mu;
    W_old   = W;
    sig_old = sig;
    
    % 1 step
    % E
    [post1, ll1] = ppca_md_E(x1(:,do1), mu(:,do1), W(do1,:,:), sig, pp);
    [post2, ll2] = ppca_md_E(x2(:,do2), mu(:,do2), W(do2,:,:), sig, pp);
    ll = ll1+ll2;
        
    % check convergence
    lldiff = ll - ll_old;
    if (lldiff >= 0 && lldiff < tolFun*abs(ll))
        break;
    end
    ll_old = ll;
    ll_hist(iter) = ll;
    
    % unnormalized component probability
    pp1 = sum(post1);
    pp2 = sum(post2);
    pp = pp1+pp2;
    
    %%%
    pp0 = find(pp==0);
    if ~isempty(pp0)
        fprintf(['iter : ' num2str(iter) ' no elements in comp : ' num2str(pp0) '\n'])
        post1(:,pp0) = realmin;
        post2(:,pp0) = realmin;
        pp1(pp0) = sum(post1(:,pp0));
        pp2(pp0) = sum(post2(:,pp0));
        pp(pp0) = pp1(pp0) + pp2(pp0);
    end
    
    %
    for k=1:K
        [x1hat, Q1] = util_xhatQ(x1, mu(k,:), W(:,:,k), sig(k), do1, dm1);
        [x2hat, Q2] = util_xhatQ(x2, mu(k,:), W(:,:,k), sig(k), do2, dm2);
        
        % mu
        mu(k,:) = (post1(:,k)'*x1hat + post2(:,k)'*x2hat) / pp(k);
        
        % W and sig
        x1hat0 = bsxfun(@minus, x1hat, mu(k,:));
        x2hat0 = bsxfun(@minus, x2hat, mu(k,:));
        
        SW = x1hat0'*(repmat(post1(:,k),1,q).*(x1hat0*W(:,:,k))) ...
           + x2hat0'*(repmat(post2(:,k),1,q).*(x2hat0*W(:,:,k)));
        SW(dm1,:) = SW(dm1,:) + pp1(k)*Q1*W(dm1,:,k);
        SW(dm2,:) = SW(dm2,:) + pp2(k)*Q2*W(dm2,:,k);
        
        trS = post1(:,k)'*sum(x1hat0.^2,2) + post2(:,k)'*sum(x2hat0.^2,2) ...
            + pp1(k)*trace(Q1) + pp2(k)*trace(Q2);
        
        Minv = inv(sig(k)*eye(q) + W(:,:,k)'*W(:,:,k));
        ttdash = pp(k)*sig(k)*eye(q) + Minv*(W(:,:,k)'*SW);
        
        W(:,:,k) = SW*inv(ttdash);
%         W(:,:,k) = SW/ttdash;
        sig(k) = (trS - trace(W(:,:,k)'*SW*Minv))/(pp(k)*d);
        
        %%%
        if sig(k)<=0
            sig(k) = sig_old(k);
        end
    end

    % normalized component probability
    pp = pp / sum(pp);
end

% component covariance
C = zeros(d,d,K);
M = zeros(q,q,K);
for k=1:K
    C(:,:,k) = sig(k)*eye(d) + W(:,:,k)*W(:,:,k)';
    M(:,:,k) = sig(k)*eye(q) + W(:,:,k)'*W(:,:,k);
end

% # of params
nParam = K-1 + K*(d + d*q + 1 - q*(q-1)/2);

% output
param.prior = prior;
param.K = K;
param.pp = pp;
param.mu = mu;
param.W = W;
param.sigma = sig;

param.C = C;
param.M = M;

param.post1 = post1;
param.post2 = post2;

param.iters = iter;
param.log_lh = ll;
param.AIC = -2*ll + 2*nParam;
param.BIC = -2*ll + nParam*log(N1+N2);
param.log_lhdiff = lldiff;
param.ll_hist = ll_hist(1:iter);

end % function ppca_md


function [pp, mu, W, sig] = util_init(x1, x2, feature, K, q, prior, init)

[N1,d] = size(x1);
N2 = size(x2,1);

% file structure
dp = feature(:,1);  % common
dq = feature(:,2);  % specific to x1
dr = feature(:,3);  % specific to x2
do1 = dp|dq;
do2 = dp|dr;

% % init
if nargin < 7 || isempty(init)
    
    % mu
    if isempty(prior)
        x1(:,dr) = repmat(mean(x2(:,dr)), N1, 1);
        x2(:,dq) = repmat(mean(x1(:,dq)), N2, 1);
        muidx = randint(K,1,[1 N1+N2]);
        mu = [x1(muidx(muidx<=N1),:); x2(muidx(muidx>N1)-N1,:)];
    else
        % prior
        mu_mean = prior.mu_mean;
        
        mu = zeros(K, d);
        for k=1:K
            mu(k,:) = mu_mean(k,:);
        end
    end
    
    % W, sig, pp
    W = zeros(d, q, K);
    sig = zeros(K, 1);
    pp = zeros(1, K);
    
    [temp, x1cls] = min(util_dist2(x1(:,do1), mu(:,do1)), [], 2);
    [temp, x2cls] = min(util_dist2(x2(:,do2), mu(:,do2)), [], 2);

    for k=1:K
        C = randn(d,d);
        C = (C+C')/2;
        
        idx1 = false(N1,1);
        idx2 = false(N2,1);
        
        idx1 = idx1 | (x1cls==k);
        idx2 = idx2 | (x2cls==k);
        
        if any(idx1)
            C(do1,do1) = cov(x1(idx1,do1));
        end
        if any(idx2)
            C(do2,do2) = cov(x2(idx2,do2));
        end
        if any(idx1) && any(idx2)
            C(dp, dp ) = cov([x1(idx1,dp); x2(idx2,dp)]);
        end
        
        % % test %%%%%%
%         C(dq, dr) = C(dq, dp)/C(dp, dp)*C(dp, dr);
%         C(dr, dq) = C(dq, dr)';
        
        [V,D] = eig(C);
        D = diag(D);
        t = D < eps(max(D))*length(D);
        if any(t)
            D(t) = min(D(~t))*0.1;
            C = V*diag(D)*V';
        end
        
        [Wk, sigk] = util_eigs(C, q);
        W(:,:,k) = repmat(Wk, [1 1 1]);
        sig(k) = sigk;
        
        pp(k) = (sum(idx1)+sum(idx2));
    end
    pp = pp/(N1+N2);
else
    pp = init.pp;
    mu = init.mu;
    W = init.W;
    sig = init.sigma;
end

end % function util_init


function [W, sig] = util_eigs(C, q)

d = size(C,1);

[V,D] = eig(C);
D = diag(D);
[D,IX] = sort(D,'descend');
V = V(:,IX);

sig = sum(D(q+1:end))/(d-q);
W = V(:,1:q) * diag(sqrt(D(1:q) - sig));

end


function [xhat, Q] = util_xhatQ(x, mu, W, sig, do, dm)

[d,q] = size(W);

% C
C = sig*eye(d) + W*W';

% CooInv
Moo = sig*eye(q) + W(do,:)'*W(do,:);
[R,f] = chol(Moo);
WRinv = W(do,:)/R;
CooInv = (eye(sum(do)) - WRinv*WRinv')/sig;
% CooInv = inv(C(do,do));

% xhat
% xhat0 = bsxfun(@minus, x, mu);
xhat0 = zeros(size(x));
xhat0(:,do) = bsxfun(@minus, x(:,do), mu(:,do));
xhat0(:,dm) = (C(dm,do)*CooInv*xhat0(:,do)')';
xhat = bsxfun(@plus, xhat0, mu);

% Q
Q = C(dm,dm) - C(dm,do)*CooInv*C(do,dm);

end % function util_xhatQ

